/************************************************************************
* Copyright (C) 2001 Thomas Schulz					*
*									*/
	static char rcsid[] =
	"$Id: ic35mgr.c,v 1.13 2001/02/03 23:06:01 tsch Rel $";  	/*
*									*
* IC35 manager								*
*									*
*************************************************************************
*									*
*	??? is "fixme" mark: sections of code needing fixes		*
*									*
*	for usage run  ic35mgr -h  or see function usage() below.	*
*									*
************************************************************************/

#include <stdio.h>	/* printf(), .. 	*/
#include <stdlib.h>	/* malloc(), free()	*/
#include <string.h>	/* memcpy(), strcpy() ..*/
#include <ctype.h>	/* islower(), ..	*/
#include <unistd.h>	/* getopt()		*/
#include <time.h>	/* struct tm, time() ..	*/
#include <utime.h>	/* struct utimbuf, ..	*/
#include <errno.h>

#include "util.h"	/* FALSE, ..		*/
#include "comio.h"	/* com_siminit()	*/
#include "mgrtrans.h"	/* IC35mgr transactions */
NOTUSED(rcsid)


extern char *	pkgvers;	/* these are all		*/
extern char *	pkgdate;	/*  in versinfo.c, which is	*/
extern char *	bldinfo;	/*   auto-generated by Makefile	*/


/* ==================================== */
/*	IC35 manager commands		*/
/* ==================================== */

/* report IC35 status
*  ------------------
*/
static int
ic35mgrstat( char * devname, char * statfname )
{
    int 	mmcnum;
    int 	rval;
    char	label[11+1];

    if ( (rval = mconnect( devname, statfname )) != OK )
	return rval;

    for ( mmcnum = 1; mmcnum <= 2; ++mmcnum ) {
	if ( (rval = mmc_status( mmcnum )) < 0 )
	    break;
	if ( rval != 0 ) {
	    if ( (rval = mmc_label( mmcnum, label )) < 0 )
		break;
	    message( "MMCard%d found, label: \"%s\"", mmcnum, label );
	} else {
	    message( "MMCard%d not present", mmcnum );
	}
    }

    mdisconnect();
    return rval;
}


/* backup IC35 database to file
*  ----------------------------
*/
static int
ic35backup( char * devname, char * fname )
{
    int 	rval;

    if ( (rval = mconnect( devname, NULL )) != OK )
	return rval;

    rval = readdatabase( fname );

    mdisconnect();
    return rval;
}

/* restore IC35 database from file
*  -------------------------------
*/
static int
ic35restore( char * devname, char * fname )
{
    int 	rval;

    if ( (rval = mconnect( devname, NULL )) != OK )
	return rval;

    rval = writedatabase( fname );

    mdisconnect();
    return rval;
}


/* list IC35 MMCard(s) directories
*  -------------------------------
*	list looks like:
*	MMCardn "Label"
*	      size  filetimestamp       attr  filepath
*	nnnnnnnnnn  yyyy-mm-dd hh:mm:ss  xxx  MMCard1\subdir\filename.ext
*/
static int
mmcdirlist( char * dirpath )
{
    MMCDIR *	dirp;
    MMCDIRENT *	dirent;
    size_t	pfxlen;
    char *	subdir;
    char	attrtext[4+1];
    int 	rval, rc;

    if ( (dirp = mmc_opendir( dirpath )) == NULL )
	return ERR;
    pfxlen = strlen( dirpath ) + 1;
    rval = OK;
    while ( (dirent = mmc_readdir( dirp )) != NULL ) {
	switch ( dirent->attr ) {
	case MMCattrDirectory:	strcpy( attrtext, "dir" );	break;
	case MMCattrArchive:	strcpy( attrtext, "arc" );	break;
	default:	    sprintf( attrtext, "0x%02X", dirent->attr & 0xFF );
	}
	printf( "%10ld  %s  %4s  %s\\%s\n",
		dirent->size, mmctstampstr( dirent->tstamp ),
		attrtext, dirpath, dirent->name );
	fflush( stdout );
	if ( dirent->attr & MMCattrDirectory
	  && strcmp( dirent->name, ".." ) != 0
	  && (subdir = malloc( pfxlen + strlen(dirent->name) + 1 )) != NULL ) {
	    sprintf( subdir, "%s\\%s", dirpath, dirent->name );
	    rc = mmcdirlist( subdir );
	    if ( rval == OK ) rval = rc;
	    free( subdir );
	}
    }
    rc = mmc_closedir( dirp );
    if ( rval == OK ) rval = rc;
    return rval;		/* errcode of first failure  or  success OK */
}
static int
ic35mmcdir( char * devname )
{
    int 	mmcnum;
    int 	rval;
    char	label[11+1];
    char	dirpath[7+1];

    if ( (rval = mconnect( devname, NULL )) != OK )
	return rval;

    for ( mmcnum = 1; mmcnum <= 2; ++mmcnum ) {
	if ( (rval = mmc_status( mmcnum )) < 0 )
	    break;
	if ( rval == 0 ) {
	    message( "MMCard%d not present", mmcnum );
	    continue;
	}
	if ( (rval = mmc_label( 1, label )) < 0 )
	    break;
	printf( "MMCard%d \"%s\"\n", mmcnum, label );
	fflush( stdout );
	sprintf( dirpath, "MMCard%d", mmcnum );
	if ( (rval = mmcdirlist( dirpath )) < 0 )
	    break;
    }

    mdisconnect();
    return rval;
}


/* ==================================== */
/*	IC35 MMCard file commands	*/
/* ==================================== */

#define MINBLKSZ	     1	/* min.blocksize			*/
#define MAXBLKSZ	0x7FFF	/* max.blocksize must fit in 15 bits	*/

static int	blocksize;	/* read/write buffer size mmcget/mmcput */

/* convert MMCard path to IC35 format & get MMCard number
*  ------------------------------------------------------
*	the ic35path will be translated to match IC35 requirements:
*	translate '/' to '\' dir separators and lowercase to upper.
*	the translation is done in place, i.e. the argument string
*	is modified.
*/
static char *	MMCard = "MMCard";

static int
_ic35mmcpath( char * ic35path )
{
    int 	mmcnum;
    char *	ptr;

		/* check/set "MMCard" prefix in ic35fpath		*/
    if ( strncasecmp( ic35path, MMCard, strlen(MMCard) ) != 0
      || ! isdigit( ic35path[strlen(MMCard)] ) )
	return ERR;
    memcpy( ic35path, MMCard, strlen(MMCard) );
    mmcnum = ic35path[strlen(MMCard)] - '0';
		/* convert ic35path '/' to '\\', lowercase to uppercase */
    for ( ptr = ic35path+strlen(MMCard); *ptr; ++ptr ) {
	if ( islower( *ptr ) )
	    *ptr = (char)toupper( *ptr );
	else if ( *ptr == '/' )
	    *ptr = '\\';
    }
    return mmcnum;
}

/* get filename from IC35 MMCard path
*  ----------------------------------
*	derive local filename from base filename in ic35path, i.e. leading
*	directory components removed, and translate to lowercase.
*	the buffer for the local filename is strdup()'d and shall be freed
*	by caller.
*/
static char *
_dupic35fname( char * ic35path )
{
    char *	ptr;
    char *	fname;

    if ( (ptr = strrchr( ic35path, '\\' )) != NULL )
	++ptr;					/* skip path's last '\' */
    else
	ptr = ic35path;
    if ( (fname = strdup( ptr )) != NULL ) {
	for ( ptr = fname; *ptr; ++ptr )
	    *ptr = tolower( *ptr );		/* lowcase local fname	*/
    }					/* no memory for local filename */
    return fname;
}

/* read file from IC35 MMCard
*  --------------------------
*	the ic35path will be translated to match IC35 requirements:
*	translate '/' to '\' dir separators and lowercase to upper.
*	on absent local filename (NULL or empty string) derive it from
*	base filename in ic35path, i.e. leading directory components
*	removed, translated to lowercase.
*	e.g. mmcard1/Ic35\App\Reversi.APP will be translated
*	to   MMCard1\IC35\APP\REVERSI.APP and local file reversi.app
*	the local file's timestamp will be set to the IC35 file's.
*	(but IC35 seems to errorneously set the IC35 file's timestamp
*	to the time of read, although the IC35 file is not modified!?)
*/
static int
ic35mmcget( char * devname, char * ic35path, char * a_fname )
{
    int 		rbuffsize = 5 * 1024;
    int 		mmcnum;
    char *		fname;
    MMCFILE *		mmcfp;
    MMCDIRENT * 	mmcfstat;
    FILE *		fp;
    struct utimbuf	ftime;
    uchar *		buff;
    size_t		blen;
    int 		rlen;
    ulong		frlen;
    int 		rval;

    if ( MINBLKSZ <= blocksize && blocksize <= MAXBLKSZ )
	rbuffsize = blocksize;
    else if ( blocksize != 0 )
	error( "bad buffer size %d, using default %d", blocksize, rbuffsize );

		/* check/set "MMCard" prefix in ic35fpath and		*/
		/* convert '/' to '\\', lowercase to uppercase		*/
    if ( (mmcnum = _ic35mmcpath( ic35path )) < 0 ) {
	error( "missing \"%s\" prefix in IC35 filepath: %s", MMCard, ic35path );
	return ERR;
    }
		/* create fname from basename(ic35path) if absent	*/
    if ( a_fname && *a_fname ) {
	fname = a_fname;
    } else
	if ( (fname = _dupic35fname( ic35path )) == NULL ) {
	    error( "no memory for local filename from %s", ic35path );
	    return ERR;
	}

		/* connect to IC35 in SyncStation via serial device	*/
    if ( (rval = mconnect( devname, NULL )) != OK )
	return rval;
		/* get MMCard status (else other MMCard ops would fail)	*/
    if ( (rval = mmc_status( mmcnum )) <= 0 ) {
	if ( rval == 0 )
	    message( "MMCard%d not present", mmcnum );
	if ( fname != a_fname )
	    free( fname );
	mdisconnect();
	return ERR;
    }

		/* read file from IC35 MMCard				*/
    if ( (mmcfp = mmc_openfile( ic35path, MMCopenexist )) != NULL ) {
	if ( (mmcfstat = mmc_statfile( mmcfp )) != NULL ) {
	    if ( (buff = malloc( blen = rbuffsize )) != NULL ) {
		if ( (fp = fopen( fname, "w" )) != NULL ) {
		    rval = OK;
		    frlen = 0;
		    fprintf( stderr, "\rread %s -> %s  ", ic35path, fname );
		    while ( (rlen = mmc_readfile( mmcfp, buff, blen )) > 0 ) {
			frlen += rlen;
			fprintf( stderr, "\rread %s -> %s %ldk  ",
					ic35path, fname, (frlen+511)/1024 );
			if ( fwrite( buff, 1, rlen, fp ) != rlen ) {
			    error( "write error %s: %s (%d)",
				    fname, strerror(errno), errno );
			    rval = ERR;
			    break;
			}
		    }
		    fclose( fp );
		    if ( rlen == 0 && rval == OK ) {
			ftime.actime  = time( NULL );
			ftime.modtime = mmctstampunixtime( mmcfstat->tstamp );
			utime( fname, &ftime ); /* set file modtime from IC35 */
			fprintf( stderr, "\n" );
		    } else /* mmc_readfile() or fwrite() failed */
			rval = ERR;
		} else				/* (fp = fopen()) == NULL */
		    error( "cannot open %s: %s (%d)",
			    fname, strerror(errno), errno );
		free( buff );
	    } else				/* (buff = malloc()) == NULL */
		error( "no memory for read buffer (%d)", rbuffsize );
	} else					/* mmc_statfile() failed */
	    rval = ERR;
	mmc_closefile( mmcfp );
    } else					/* mmc_openfile() failed */
	rval = ERR;

		/* cleanup and disconnect from IC35			*/
    if ( fname != a_fname )
	free( fname );
    mdisconnect();
    return rval;
}

/* write file to IC35 MMCard
*  -------------------------
*	see ic35mmcget() comment about ic35path conversions and deriving
*	fname from it if a_fname is NULL or empty.
*/
static int
ic35mmcput( char * devname, char * ic35path, char * a_fname )
{
    int 		wbuffsize = 2 * 1024;
    int 		mmcnum;
    char *		fname;
    MMCFILE *		mmcfp;
    FILE *		fp;
    uchar *		buff;
    size_t		blen;
    int 		rlen;
    ulong		fwlen;
    int 		rval;

    if ( MINBLKSZ <= blocksize && blocksize <= MAXBLKSZ )
	wbuffsize = blocksize;
    else if ( blocksize != 0 )
	error( "bad buffer size %d, using default %d", blocksize, wbuffsize );

		/* check/set "MMCard" prefix in ic35fpath and		*/
		/* convert '/' to '\\', lowercase to uppercase		*/
    if ( (mmcnum = _ic35mmcpath( ic35path )) < 0 ) {
	error( "missing \"%s\" prefix in IC35 filepath: %s", MMCard, ic35path );
	return ERR;
    }
		/* create fname from basename(ic35path) if absent	*/
    if ( a_fname && *a_fname ) {
	fname = a_fname;
    } else
	if ( (fname = _dupic35fname( ic35path )) == NULL ) {
	    error( "no memory for local filename from %s", ic35path );
	    return ERR;
	}

		/* connect to IC35 in SyncStation via serial device	*/
    if ( (rval = mconnect( devname, NULL )) != OK )
	return rval;
		/* get MMCard status (else other MMCard ops would fail)	*/
    if ( (rval = mmc_status( mmcnum )) <= 0 ) {
	if ( rval == 0 )
	    message( "MMCard%d not present", mmcnum );
	if ( fname != a_fname )
	    free( fname );
	mdisconnect();
	return ERR;
    }

		/* write file to IC35 MMCard				*/
    if ( (mmcfp = mmc_openfile( ic35path, MMCcreatrunc )) != NULL ) {
	if ( (buff = malloc( blen = wbuffsize )) != NULL ) {
	    if ( (fp = fopen( fname, "r" )) != NULL ) {
		rval = OK;
		fwlen = 0;
		fprintf( stderr, "\rwrite %s <- %s  ", ic35path, fname );
		while ( (rlen = fread( buff, 1, blen, fp )) > 0 ) {
		    if ( mmc_writefile( mmcfp, buff, rlen ) != rlen ) {
			rval = ERR;
			break;
		    }
		    fwlen += rlen;
		    fprintf( stderr, "\rwrite %s <- %s %ldk  ",
				    ic35path, fname, (fwlen+511)/1024 );
		}
		if ( rlen <= 0 && feof( fp ) && rval == OK ) {
		    fprintf( stderr, "\n" );
		} else {		/* fread() or mmc_writefile() failed */
		    if ( rval != ERR )
			error( "read error %s: %s (%d)",
				fname, strerror(errno), errno );
		    rval = ERR;
		}
		fclose( fp );
	    } else				/* (fp = fopen()) == NULL */
		error( "cannot open %s: %s (%d)",
			fname, strerror(errno), errno );
	    free( buff );
	} else					/* (buff = malloc()) == NULL */
	    error( "no memory for write buffer (%d)", wbuffsize );
	mmc_closefile( mmcfp );
    } else					/* mmc_openfile() failed */
	rval = ERR;

		/* cleanup and disconnect from IC35			*/
    if ( fname != a_fname )
	free( fname );
    mdisconnect();
    return rval;
}

/* delete file on IC35 MMCard
*  -------------------------
*/
static int
ic35mmcdel( char * devname, char * ic35path )
{
    int 	mmcnum;
    int 	rval;

		/* check/set "MMCard" prefix in ic35fpath and		*/
		/* convert '/' to '\\', lowercase to uppercase		*/
    if ( (mmcnum = _ic35mmcpath( ic35path )) < 0 ) {
	error( "missing \"%s\" prefix in IC35 filepath: %s", MMCard, ic35path );
	return ERR;
    }

		/* connect to IC35 in SyncStation via serial device	*/
    if ( (rval = mconnect( devname, NULL )) != OK )
	return rval;
		/* get MMCard status (else delete MMCard file fails)	*/
    if ( (rval = mmc_status( mmcnum )) > 0 ) {
	message( "delete %s", ic35path );
	rval = mmc_delfile( ic35path );     /* delete IC35 MMCard file	*/
    } else if ( rval == 0 ) {
	message( "MMCard%d not present", mmcnum );
	rval = ERR;
    }
		/* disconnect from IC35					*/
    mdisconnect();
    return rval;
}


/* -------------------- M A I N - program ----------------------------- */
static void
usage( void )
{
    static char * usetext[] = {
	"usage:  ic35mgr [options] command [argument..]",
	"commands and arguments are:",
	"  status [file]          report IC35 status (and write to file)",
	"  backup file            backup IC35 database to file",
	"  restore file           restore IC35 database from file",
	"  mmcdir                 list contents of IC35 MMCard(s)",
	"  mmcdel ic35path        delete MMCard file 'ic35path'",
	"  mmcget ic35path [file] read file from ic35path on IC35 MMCard",
	"  mmcput ic35path [file] write file to ic35path on IC35 MMCard",
	"               'ic35path' is e.g. MMCard1\\IC35\\APP\\REVERSI.APP",
	"               filename 'file' is taken from ic35path if absent",
	"options are:",
	" -d device     serial device where IC35 is connected, default /dev/ic35",
	" -l logfile    file for communication log, default no logging",
	" -b size       read/write block size for mmcget/mmcput",
	" -s simfile    simulate communication with IC35 using simfile",
	"               containing log processed by ic35log",
	NULL
    };
    char **	lptr;

    printf( "IC35manager for GNU/Linux %s (%s)\n", pkgvers, pkgdate );
    printf( "%s\n", bldinfo );
    printf( "Copyright (C) 2001 Thomas Schulz\n\n" );
    for ( lptr = usetext; *lptr != NULL; ++lptr )
	printf( "%s\n", *lptr );
    exit( 2 );
}

int
main( int argc, char *argv[] )
{
    char *	devname = "/dev/ic35";
    char *	command   = NULL;
    char *	logfname = NULL;
    char *	simfname = NULL;
    int 	rval;

	/*
	* parse command line
	*/
    for ( ; ; ) {
	switch ( getopt( argc, argv, "d:p:f:l:s:b:h" ) ) {
	default:
	case 'h':			/*  show Help and exit		*/
	    usage();
	    return 2;
	case 'd':			/*  serial Device where IC35	*/
	    devname = optarg;
	    continue;
	case 'l':			/*  communication logfile	*/
	    logfname = optarg;
	    continue;
	case 's':			/*  simulate: no communication	*/
	    simfname = optarg;
	    continue;
	case 'b':			/*  buffer size for mmcget,put	*/
	    blocksize = atoi( optarg );
	    continue;
	case -1:			/*  end of options		*/
	    break;
	}
	break;
    }
    if ( argc < optind+1 ) {
	error( "missing command\nuse \"ic35mgr -h\" for help" );
	return 1;
    }

	/*
	* initial info to logfile: date,time, version, command line
	* if 'logfname' is NULL or empty, nothing will be logged.
	*/
    log_init( logfname, "ic35mgr", 0 );
    log_proginfo( "ic35mgr", pkgvers, pkgdate, bldinfo );
    log_argsinfo( argc, argv );

	/*
	* setup simulated communication if requested
	*/
    if ( simfname )
	com_siminit( simfname );

	/*
	* process command
	*/
    command = argv[optind];		/* note command			*/
    argv += optind, argc -= optind;	/* skip options			*/
    rval = 1;
    if (	strcmp( command, "status"  ) == 0 ) {
	if ( argc <= 2 )
	    rval = ic35mgrstat( devname, argc > 1 ? argv[1]	/*???*/
						  : "/tmp/ic35.stat" );
	else
	    error( "too many arguments" );
    } else if ( strcmp( command, "backup"  ) == 0 ) {
	if ( argc == 2 )
	    rval = ic35backup(  devname, argv[1] );
	else
	    error( argc < 2 ? "missing backup file"
			    : "too many arguments" );
    } else if ( strcmp( command, "restore" ) == 0 ) {
	if ( argc == 2 )
	    rval = ic35restore( devname, argv[1] );
	else
	    error( argc < 2 ? "missing restore file"
			    : "too many arguments" );
    } else if ( strcmp( command, "mmcdir"  ) == 0 ) {
	if ( argc == 1 )
	    rval = ic35mmcdir( devname ) == OK ? 0 : 1;
	else
	    error( "too many arguments" );
    } else if ( strcmp( command, "mmcdel"  ) == 0 ) {
	if ( argc == 2 )
	    rval = ic35mmcdel( devname, argv[1] ) == OK ? 0 : 1;
	else
	    error( argc < 2 ? "missing ic35path"
			    : "too many arguments" );
    } else if ( strcmp( command, "mmcget"  ) == 0 ) {
	if ( argc == 2 || argc == 3 )
	    rval = ic35mmcget( devname, argv[1], argc == 3 ? argv[2] : NULL );
	else
	    error( argc < 2 ? "missing ic35path"
			    : "too many arguments" );
    } else if ( strcmp( command, "mmcput"  ) == 0 ) {
	if ( argc == 2 || argc == 3 )
	    rval = ic35mmcput( devname, argv[1], argc == 3 ? argv[2] : NULL );
	else
	    error( argc < 2 ? "missing ic35path,file"
			    : "too many arguments" );
    } else {
	error( "unsupported command: %s", command );
	rval = 1;
    }
    if ( rval != OK )
	rval = 1;

    log_close();
    return rval;
}
