/************************************************************************
* Copyright (C) 2000 Thomas Schulz					*
*									*/
	static char rcsid[] =
	"$Id: util.c,v 1.5 2001/02/03 23:10:39 tsch Rel $";  	/*
*									*
* IC35 utilities: logging, errors, messages				*
*									*
*************************************************************************
*									*
*	??? is "fixme" mark: sections of code needing fixes		*
*									*
************************************************************************/

#include <stdio.h>	/* fprintf(), ..	*/
#include <stdarg.h>	/* va_start(), ..	*/
#include <stdlib.h>	/* malloc(), free()	*/
#include <string.h>	/* strcpy(), .. 	*/
#include <sys/types.h>	/* size_t, ..		*/
#include <sys/time.h>	/* gettimeofday(), ..	*/

#include "util.h"
NOTUSED(rcsid)


/* logging
*  -------
*	format of log entries:
*	hh:mm:ss.mmmm tag level message
*	hh:mm:ss.mmmm tag level  addr hex... char...
*/

static char *	logfname = NULL;
static FILE *	logfp    = NULL;
static char *	logtag   = "";
static int	loglevel = L_DEBUG;

void
log_init( char * l_fname, char * l_tag, int level )
{
    if ( l_fname && *l_fname )
	logfp = fopen( logfname = l_fname, "w" );
    if ( l_tag && *l_tag )
	logtag = l_tag;
    if ( level )
	loglevel = level;
}
void
log_close( void )
{
    if ( logfp ) {
	fclose( logfp );
	logfp = NULL;
    }
}
static char *
ltstamp( void )
{
    static char 	timestamp[16];
    struct timeval	tv;
    struct tm *		ptm;

    gettimeofday( &tv, NULL );
    ptm = localtime( (time_t*)&tv.tv_sec );
    sprintf( timestamp, "%02d:%02d:%02d.%04d",
			ptm->tm_hour, ptm->tm_min, ptm->tm_sec,
			(int)(tv.tv_usec/100) );
    return timestamp;
}
static void
vltprintf( int level, char * tstamp, const char * format, va_list argp )
{
    if ( logfp == NULL || level > loglevel )
	return;
    fprintf( logfp, "%s %s %d ", tstamp, logtag, level );
    vfprintf( logfp, format, argp );
    fprintf( logfp, "\n" );
}
static void
vlprintf( int level, const char * format, va_list argp )
{
    vltprintf( level, ltstamp(), format, argp );
}
void
lprintf( int level, const char * format, ... )
{
    va_list	argp;

    va_start( argp, format );
    vlprintf( level, format, argp );
    va_end( argp );
}
static void
ltprintf( int level, char * tstamp, const char * format, ... )
{
    va_list	argp;

    va_start( argp, format );
    vltprintf( level, tstamp, format, argp );
    va_end( argp );
}
void
ldump( int level, void * data, size_t dlen, const char * format, ... )
{
    va_list	argp;
    char *	tstamp;
    int 	i;
    uchar	*dptr, *dend;
    char	*hdptr, hdump[5+3*16+1];
    char	*cdptr, cdump[16+1];

    tstamp = ltstamp();
    va_start( argp, format );
    vltprintf( level, tstamp, format, argp );
    va_end( argp );
    dend = (dptr = (uchar*)data) + dlen;
    while ( dptr < dend ) {
	hdptr = hdump;
	cdptr = memset( cdump, 0, sizeof(cdump) );
	sprintf( hdptr, "%04X:", dptr - (uchar*)data ); hdptr += 5;
	for ( i = 0; i < 16 && dptr < dend; ++i, ++dptr ) {
	    sprintf( hdptr, " %02X", *dptr ); hdptr += 3;
	    *cdptr++ = ' ' < *dptr && *dptr <= '~' ? *dptr : '.';
	}
	ltprintf( level, tstamp, " %-53s %s", hdump, cdump );
    }
}

/* output (error-)message
*  ----------------------
*/
static void
vlmessage( int level, const char * format, va_list argp )
{
    vfprintf( stderr, format, argp );
    fprintf( stderr, "\n" );
    vlprintf( level, format, argp );
}
void
error( const char * format, ... )
{
    va_list	argp;

    va_start( argp, format );
    vlmessage( L_ERROR, format, argp );
    va_end( argp );
}
void
message( const char * format, ... )
{
    va_list	argp;

    va_start( argp, format );
    vlmessage( L_MESG, format, argp );
    va_end( argp );
}

/* report unimplemented feature
*  ----------------------------
*/
int
_not_impl( char * feature )
{
    error( "%s NOT IMPLEMENTED", feature );
    return ERR;
}

/* log program info
*  ----------------
*	current date and time, as log has only hh:mm:ss.mmmm stamps
*	program name, version and release date
*	build date+time, user, host
*/
void
log_proginfo( char * name, char * version, char * date, char * bldinfo )
{
    time_t	tnow;
    struct tm *	ptm;
    char	dtime[24+1];	/* yyyy-mm-dd hh:mm:ss zzzz */

    tnow = time( NULL );
    ptm = localtime( &tnow );
    strftime( dtime, sizeof(dtime), "%Y-%m-%d %T %Z", ptm );
    lprintf( L_MESG, "%s", dtime );
    lprintf( L_MESG, "%s %s (%s)", name, version, date );
    lprintf( L_MESG, "%s", bldinfo );
}

/* log command line
*  ----------------
*/
void
log_argsinfo( int argc, char ** argv )
{
    int 	i, len;
    char *	cmdline;

    for ( i = len = 0; i < argc; ++i )
	len += strlen( argv[i] ) + 1;
    if ( (cmdline = malloc( len )) != NULL ) {
	strcpy( cmdline, argv[0] );
	for ( i = 1; i < argc; ++i )
	    strcat( strcat( cmdline, " " ), argv[i] );
	lprintf( L_MESG, "cmd: %s", cmdline );
	free( cmdline );
    }
}
